@@ -9,7 +9,6 @@
 
 #include <stddef.h>
 #include <limits.h>
-#include "string.hpp"
 #include "mutable_string.hpp"
 
 namespace acommon {
@@ -34,11 +33,6 @@ namespace acommon {
     char * e = unescape(d.str, d.str);
     d.size = e - d.str;
   }
-  static inline void unescape(String & d) {
-    d.ensure_null_end();
-    char * e = unescape(d.data(), d.data());
-    d.resize(e - d.data());
-  }
 
   // if limit is not given than dest should have enough space for 
   // twice the number of characters of src
@@ -46,12 +40,11 @@ namespace acommon {
 	      size_t limit = INT_MAX, const char * others = 0);
 
   void to_lower(char *);
-  void to_lower(String &, const char *);
 
-  // extract the first whitespace delimited word from d.value and put
+  // extract the first whitespace delemited word from d.value and put
   // it in d.key.  d.value is expected not to have any leading
-  // whitespace. The string is modified in place so that d.key will be
-  // null terminated.  d.value is modified to point to any additional
+  // whitespace. The string is modifed in place so that d.key will be
+  // null terminated.  d.value is modifed to point to any additional
   // data after key with any leading white space removed.  Returns
   // false when there are no more words to extract
   bool split(DataPair & d);
@@ -61,8 +54,5 @@ namespace acommon {
 
   bool getline(IStream & in, DataPair & d, String & buf);
 
-  char * get_nb_line(IStream & in, String & buf);
-  void remove_comments(String & buf);
-
 }
 #endif
