@@ -18,19 +18,6 @@
 #include "ostream.hpp"
 #include "istream.hpp"
 
-//
-// acommon::String is similar to std::string, but without many of the
-// extra non-stl like methods.  The string is guaranteed to be stored
-// in a continues areas of memory but is not guaranteed to be null
-// terminated.  However, space is always allocated for the null
-// characters.  Thus, the c_str() method will never invalided any
-// exiting pointers.  The string is also null terminated when accessed
-// via the str() and mstr() methods.  In addition the method
-// ensure_null_end() will null terminate the string.  Once null
-// terminated the string will remain as such until the length of the
-// string changes.
-//
-
 namespace acommon {
 
   template <typename Ret> class PosibErr;
@@ -47,30 +34,13 @@ namespace acommon {
     char * begin_;
     char * end_;
     char * storage_end_;
-
-    void assign_only_nonnull(const char * b, unsigned size)
+    void assign_only(const char * b, size_t size) 
     {
       begin_ = (char *)malloc(size + 1);
-      memmove(begin_, b, size);
+      memcpy(begin_, b, size);
       end_   = begin_ + size;
       storage_end_ = end_ + 1;
     }
-    void zero() 
-    {
-      begin_ = 0;
-      end_ = 0;
-      storage_end_ = 0;
-    }
-    void assign_only(const char * b)
-    {
-      if (b && *b) assign_only_nonnull(b, strlen(b));
-      else zero();
-    }
-    void assign_only(const char * b, unsigned size) 
-    {
-      if (b && size > 0) assign_only_nonnull(b, size);
-      else zero();
-    }
     void reserve_i(size_t s = 0);
   public:
     void reserve(size_t s) 
@@ -96,10 +66,7 @@ namespace acommon {
     size_t max_size() const {return INT_MAX;}
     size_t capacity() const {return storage_end_ ? storage_end_ - begin_ - 1 : 0;}
 
-    void ensure_null_end() const {
-      if (!begin_) const_cast<String *>(this)->reserve_i();
-      *end_ = '\0';
-    }
+    void ensure_null_end() const {*end_ = '\0';}
 
     const char * c_str() const {
       if (begin_) {ensure_null_end(); return begin_;}
@@ -129,10 +96,10 @@ namespace acommon {
     }
 
     char & operator[] (size_t pos) {return begin_[pos];}
-    char operator[] (size_t pos) const {return begin_[pos];}
+    const char operator[] (size_t pos) const {return begin_[pos];}
 
     char & back() {return end_[-1];}
-    char back() const {return end_[-1];}
+    const char back() const {return end_[-1];}
 
     void clear() {end_ = begin_;}
 
@@ -141,11 +108,15 @@ namespace acommon {
     //
 
     String() : begin_(0), end_(0), storage_end_(0) {}
-    String(const char * s) {assign_only(s);}
-    String(const char * s, unsigned size) {assign_only(s, size);}
-    String(ParmStr s) {assign_only(s, s.size());}
+    String(const char * s) {assign_only(s, strlen(s));}
+    String(const char * s, unsigned int size) {assign_only(s, size);}
+    String(ParmString s) {assign_only(s, s.size());}
     String(MutableString s) {assign_only(s.str, s.size);}
     String(const String & other) {assign_only(other.begin_, other.end_-other.begin_);}
+#ifndef __SUNPRO_CC
+    // This causes a conflict with the copy constructor on Suns comp
+    inline String(const PosibErr<String> & other);
+#endif
 
     //
     // assign
@@ -156,20 +127,20 @@ namespace acommon {
       clear();
       if (size != 0) {
         reserve(size);
-        memmove(begin_, b, size);
+        memcpy(begin_, b, size);
         end_   = begin_ + size;
       } 
     }
     void assign(const char * b) 
     {
-      if (b) assign(b, strlen(b));
+      assign(b, strlen(b));
     }
     String & operator= (const char * s) {
       assign(s);
       return *this;
     }
     inline String & operator= (const PosibErr<const char *> & s);
-    String & operator= (ParmStr s) {
+    String & operator= (ParmString s) {
       assign(s, s.size());
       return *this;
     }
@@ -190,7 +161,7 @@ namespace acommon {
     String & append(const void * str, unsigned int sz)
     {
       reserve(size() + sz);
-      if (sz > 0) memcpy(end_, str, sz);
+      memcpy(end_, str, sz);
       end_ += sz;
       return *this;
     }
@@ -228,7 +199,7 @@ namespace acommon {
       append(c);
       return *this;
     }
-    String & operator+= (ParmStr s) {
+    String & operator+= (ParmString s) {
       if (s.have_size())
         append(s, s.size());
       else
@@ -321,7 +292,7 @@ namespace acommon {
       erase(begin_ + pos, begin_ + pos + s);
     }
 
-    //FIXME: Make this more efficient by rewriting the implementation
+    //FIXME: Make this more efficent by rewriting the implemenation
     //       to work with raw memory rather than using vector<char>
     template <typename Itr>
     void replace(iterator start, iterator stop, Itr rstart, Itr rstop) 
@@ -353,12 +324,12 @@ namespace acommon {
       return pos;
     }
 
-    bool prefix(ParmStr str, size_t offset = 0) const
+    bool prefix(ParmString str, size_t offset) const
     {
       if (str.size() > size() - offset) return false;
       return memcmp(begin_ + offset, str.str(), str.size()) == 0;
     };
-    bool suffix(ParmStr str) const
+    bool suffix(ParmString str) const
     {
       if (str.size() > size()) return false;
       return memcmp(end_ - str.size(), str.str(), str.size()) == 0;
@@ -366,6 +337,7 @@ namespace acommon {
 
     // FIXME: Eventually remove
     static const size_t npos = INT_MAX;
+    size_t length() const {return size();}
     size_t find(char c, size_t pos = 0) const {
       char * res = (char *)memchr(begin_ + pos, c, size() - pos);
       if (res == 0) return npos;
@@ -392,11 +364,11 @@ namespace acommon {
       {return reinterpret_cast<unsigned int &>(operator[](pos));}
 
     void write (char c) {append(c);}
-    void write (ParmStr str) {operator+=(str);}
+    void write (ParmString str) {operator+=(str);}
     void write (const void * str, unsigned int sz) {append(str,sz);}
 
 
-    String & operator << (ParmStr str) {
+    String & operator << (ParmString str) {
       append(str);
       return *this;
     }
@@ -407,68 +379,22 @@ namespace acommon {
     }
   };
 
-  inline String operator+ (ParmStr lhs, ParmStr rhs)
+  inline String operator+ (ParmString rhs, ParmString lhs)
   {
     String tmp;
-    tmp.reserve(lhs.size() + rhs.size());
-    tmp += lhs;
+    tmp.reserve(rhs.size() + lhs.size());
     tmp += rhs;
+    tmp += lhs;
     return tmp;
   }
 
-  inline bool operator== (const String & x, const String & y)
-  {
-    if (x.size() != y.size()) return false;
-    if (x.size() == 0) return true;
-    return memcmp(x.data(), y.data(), x.size()) == 0;
-  }
-  inline bool operator== (const String & x, const char * y)
-  {
-    return strcmp(x.c_str(), y) == 0;
-  }
-  inline bool operator== (const char * x, const String & y)
-  {
-    return strcmp(x, y.c_str()) == 0;
-  }
-  inline bool operator== (const String & x, ParmStr y)
-  {
-    if (y == 0) return x.size() == 0;
-    return strcmp(x.c_str(), y) == 0;
-  }
-  inline bool operator== (ParmStr x, const String & y)
-  {
-    if (x == 0) return y.size() == 0;
-    return strcmp(x, y.c_str()) == 0;
-  }
-
-  inline bool operator!= (const String & x, const String & y)
-  {
-    return !(x == y);
-  }
-  inline bool operator!= (const String & x, const char * y)
-  {
-    return strcmp(x.c_str(), y) != 0;
-  }
-  inline bool operator!= (const char * x, const String & y)
-  {
-    return strcmp(x, y.c_str()) != 0;
-  }
-  inline bool operator!= (const String & x, ParmStr y)
-  {
-    return !(x == y);
-  }
-  inline bool operator!= (ParmStr x, const String & y)
-  {
-    return !(x == y);
-  }
-
   inline ParmString::ParmString(const String & s) : str_(s.c_str()), size_(s.size()) {}
 
   class StringIStream : public IStream {
     const char * in_str;
     char         delem;
   public:
-    StringIStream(ParmStr s, char d = ';')
+    StringIStream(ParmString s, char d = ';')
       : IStream(d), in_str(s) {}
     bool append_line(String & str, char c);
     bool read(void * data, unsigned int size);
@@ -492,7 +418,7 @@ namespace acommon {
 
 namespace std
 {
-  template<> inline void swap(acommon::String & x, acommon::String & y) {return x.swap(y);}
+  template<> static inline void swap(acommon::String & x, acommon::String & y) {return x.swap(y);}
 }
 
 #endif
