@@ -45,7 +45,7 @@
 #include <cstring>
 #include <cstdio>
 
-//#include "iostream.hpp"
+#include "iostream.hpp"
 
 #include "affix.hpp"
 #include "errors.hpp"
@@ -55,9 +55,6 @@
 #include "speller_impl.hpp"
 #include "vararray.hpp"
 #include "lsort.hpp"
-#include "hash-t.hpp"
-
-#include "gettext.h"
 
 using namespace std;
 
@@ -68,47 +65,38 @@ static char EMPTY[1] = {0};
 
 //////////////////////////////////////////////////////////////////////
 //
-// Entry struct definations
+//*Entry struct definations
 //
 
-struct Conds
-{
-  char * str;
-  unsigned num;
-  char conds[SETSIZE];
-  char get(byte i) const {return conds[i];}
-};
-
 struct AffEntry
 {
-  const char *   appnd;
-  const char *   strip;
-  byte           appndl;
-  byte           stripl;
-  byte           xpflg;
-  char           achar;
-  const Conds *  conds;
-  //unsigned int numconds;
-  //char         conds[SETSIZE];
+  char *         appnd;
+  char *         strip;
+  unsigned int   appndl;
+  unsigned int   stripl;
+  unsigned int   numconds;
+  int            xpflg;
+  char         achar;
+  char         conds[SETSIZE];
 };
 
 // A Prefix Entry
   
 struct PfxEntry : public AffEntry
 {
+  AffixMgr * pmyMgr;
+  
   PfxEntry * next;
   PfxEntry * next_eq;
   PfxEntry * next_ne;
   PfxEntry * flag_next;
-  PfxEntry() {}
+  PfxEntry(AffixMgr * pmgr) : pmyMgr(pmgr) {}
 
-  bool check(const LookupInfo &, const AffixMgr * pmyMgr,
-             ParmString, CheckInfo &, GuessInfo *, bool cross = true) const;
+  bool check(const LookupInfo &, ParmString, CheckInfo &, GuessInfo *) const;
 
   inline bool          allow_cross() const { return ((xpflg & XPRODUCT) != 0); }
   inline byte flag() const { return achar;  }
   inline const char *  key() const  { return appnd;  }
-  bool applicable(SimpleString) const;
   SimpleString add(SimpleString, ObjStack & buf) const;
 };
 
@@ -116,14 +104,15 @@ struct PfxEntry : public AffEntry
 
 struct SfxEntry : public AffEntry
 {
-  const char * rappnd; // this is set in AffixMgr::build_sfxlist
+  AffixMgr*    pmyMgr;
+  char *       rappnd; // this is set in AffixMgr::build_sfxlist
   
   SfxEntry *   next;
   SfxEntry *   next_eq;
   SfxEntry *   next_ne;
   SfxEntry *   flag_next;
 
-  SfxEntry() {}
+  SfxEntry(AffixMgr* pmgr) : pmyMgr(pmgr) {}
 
   bool check(const LookupInfo &, ParmString, CheckInfo &, GuessInfo *,
              int optflags, AffEntry * ppfx);
@@ -131,8 +120,7 @@ struct SfxEntry : public AffEntry
   inline bool          allow_cross() const { return ((xpflg & XPRODUCT) != 0); }
   inline byte flag() const { return achar;  }
   inline const char *  key() const  { return rappnd; } 
-  bool applicable(SimpleString) const;
-  SimpleString add(SimpleString, ObjStack & buf, int limit, SimpleString) const;
+  SimpleString add(SimpleString, ObjStack & buf, int limit = INT_MAX) const;
 };
 
 //////////////////////////////////////////////////////////////////////
@@ -167,76 +155,44 @@ struct AffixLess
   bool operator() (T * x, T * y) const {return strcmp(x->key(),y->key()) < 0;}
 };
 
-// struct StringLookup {
-//   struct Parms {
-//     typedef const char * Value;
-//     typedef const char * Key;
-//     static const bool is_multi = false;
-//     hash<const char *> hfun;
-//     size_t hash(const char * s) {return hfun(s);}
-//     bool equal(const char * x, const char * y) {return strcmp(x,y) == 0;}
-//     const char * key(const char * c) {return c;}
-//   };
-//   typedef HashTable<Parms> Lookup;
-//   Lookup lookup;
-//   ObjStack * data_buf;
-//   StringLookup(ObjStack * b) : data_buf(b) {}
-//   const char * dup(const char * orig) {
-//     pair<Lookup::iterator, bool> res = lookup.insert(orig);
-//     if (res.second) *res.first = data_buf->dup(orig);
-//     return *res.first;
-//     //return data_buf->dup(orig);
-//   }
-// };
+//////////////////////////////////////////////////////////////////////
+//
+// CheckList
+//
 
-struct CondsLookupParms {
-  typedef const Conds * Value;
-  typedef const char * Key;
-  static const bool is_multi = false;
-  acommon::hash<const char *> hfun;
-  size_t hash(const char * s) {return hfun(s);}
-  bool equal(const char * x, const char * y) {return strcmp(x,y) == 0;}
-  const char * key(const Conds * c) {return c->str;}
-};
-
-typedef HashTable<CondsLookupParms> CondsLookup;
-
-// normalizes and checks the cond_str
-// returns the length of the new string or -1 if invalid
-static int normalize_cond_str(char * str)
+CheckList * new_check_list()
 {
-  char * s = str;
-  char * d = str;
-  while (*s) {
-    if (*s != '[') {
-      *d++ = *s++;
-    } else if (s[1] == '\0' || s[1] == ']') {
-      return -1;
-    } else if (s[2] == ']') {
-      *d++ = s[1];
-      s += 3;
-    } else {
-      *d++ = *s++;
-      if (*s == '^') *d++ = *s++;
-      while (*s != ']') {
-        if (*s == '\0' || *s == '[') return -1;
-        char * min = s;
-        for (char * i = s + 1; *i != ']'; ++i) {
-          if ((byte)*i < (byte)*min) min = i;}
-        char c = *s;
-        *d++ = *min;
-        *min = c;
-        ++s;
-      }
-      *d++ = *s++;
-    }
-  }
-  *d = '\0';
-  return d - str;
+  return new CheckList;
 }
 
-static void encodeit(CondsLookup &, ObjStack &, 
-                     AffEntry * ptr, char * cs);
+void delete_check_list(CheckList * cl)
+{
+  delete cl;
+}
+
+CheckInfo * check_list_data(CheckList * cl)
+{
+  if (cl->gi.num > 0)
+    return cl->data + 1;
+  else
+    return 0;
+}
+
+CheckList::CheckList()
+ : gi(63)
+{
+  memset(data, 0, sizeof(data));
+  gi.reset(data);
+}
+
+void CheckList::reset()
+{
+  for (CheckInfo * p = data + 1; p != data + 1 + gi.num; ++p) {
+    free(const_cast<char *>(p->word.str()));
+    p->word = 0;
+  }
+  gi.reset(data);
+}
 
 //////////////////////////////////////////////////////////////////////
 //
@@ -258,12 +214,36 @@ PosibErr<void> AffixMgr::setup(ParmString affpath, Conv & iconv)
   return parse_file(affpath, iconv);
 }
 
-AffixMgr::AffixMgr(const Language * l) 
-  : lang(l), data_buf(1024*16) {}
+AffixMgr::~AffixMgr()
+{
+  // pass through linked prefix entries and clean up
+  for (int i=0; i < SETSIZE ;i++) {
+    pFlag[i] = NULL;
+    PfxEntry * ptr = pStart[i];
+    PfxEntry * nptr = NULL;
+    while (ptr) {
+      nptr = ptr->next;
+      delete ptr;
+      ptr = nptr;
+      nptr = NULL;
+    }
+  }
 
-AffixMgr::~AffixMgr() {}
+  // pass through linked suffix entries and clean up
+  for (int j=0; j < SETSIZE ; j++) {
+    sFlag[j] = NULL;
+    SfxEntry * ptr = sStart[j];
+    SfxEntry * nptr = NULL;
+    while (ptr) {
+      nptr = ptr->next;
+      delete ptr;
+      ptr = nptr;
+      nptr = NULL;
+    }
+  }
+}
 
-static inline void max_(int & lhs, int rhs) 
+static inline void MAX(int & lhs, int rhs) 
 {
   if (lhs < rhs) lhs = rhs;
 }
@@ -273,11 +253,9 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
 {
   // io buffers
   String buf; DataPair dp;
-
-  CondsLookup conds_lookup;
  
   // open the affix file
-  affix_file = data_buf.dup(affpath);
+  affix_file = strings.dup(affpath);
   FStream afflst;
   RET_ON_ERR(afflst.open(affpath,"r"));
 
@@ -287,8 +265,6 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
   // read in each line ignoring any that do not
   // start with a known line type indicator
 
-  char prev_aff = '\0';
-
   while (getdata_pair(afflst,dp,buf)) {
     char affix_type = ' ';
 
@@ -296,14 +272,17 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
 
     if (dp.key == "SET") {
       String buf;
-      encoding = data_buf.dup(fix_encoding_str(dp.value, buf));
+      encoding = strings.dup(fix_encoding_str(dp.value, buf));
+      char msg[96];
+      snprintf(msg, 96, _("Expected the file to be in \"%s\" not \"%s\"."),
+               lang->data_encoding(), encoding);
       if (strcmp(encoding, lang->data_encoding()) != 0)
-        return make_err(incorrect_encoding, affix_file, lang->data_encoding(), encoding);
+        return make_err(bad_file_format, affix_file, msg);
     }
 
     /* parse in the flag used by the controlled compound words */
     //else if (d.key == "COMPOUNDFLAG")
-    //  compound = data_buf.dup(d.value);
+    //  compound = strings.dup(d.value);
 
     /* parse in the flag used by the controlled compound words */
     //else if (d.key == "COMPOUNDMIN")
@@ -323,47 +302,43 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
     int numents = 0;      // number of affentry structures to parse
     char achar='\0';      // affix char identifier
     short xpflg=0;
-    AffEntry * nptr;
+    StackPtr<AffEntry> nptr;
     {
       // split affix header line into pieces
       split(dp);
       if (dp.key.empty()) goto error;
       // key is affix char
-      const char * astr = iconv(dp.key);
-      if (astr[0] == '\0' || astr[1] != '\0') goto error;
-      achar = astr[0];
-      if (achar == prev_aff) goto error_count;
-      prev_aff = achar;
+      achar = iconv(dp.key)[0];
 
       split(dp);
-      if (dp.key.size != 1 || 
-          !(dp.key[0] == 'Y' || dp.key[0] == 'N')) goto error;
+      if (dp.key.empty()) goto error;
       // key is cross product indicator 
       if (dp.key[0] == 'Y') xpflg = XPRODUCT;
     
       split(dp);
       if (dp.key.empty()) goto error;
       // key is number of affentries
-      
       numents = atoi(dp.key); 
   
       for (int j = 0; j < numents; j++) {
         getdata_pair(afflst, dp, buf);
 
-        if (affix_type == 'P') {
-          nptr = (AffEntry *) data_buf.alloc_bottom(sizeof(PfxEntry));
-          new (nptr) PfxEntry;
-        } else {
-          nptr = (AffEntry *) data_buf.alloc_bottom(sizeof(SfxEntry));
-          new (nptr) SfxEntry;
-        }
+        if (affix_type == 'P')
+          nptr.reset(new PfxEntry(this));
+        else
+          nptr.reset(new SfxEntry(this));
 
         nptr->xpflg = xpflg;
 
         split(dp);
         if (dp.key.empty()) goto error;
         // key is affix charter
-        if (iconv(dp.key)[0] != achar) goto error_count;
+        if (iconv(dp.key)[0] != achar) {
+          char msg[64];
+          snprintf(msg, 64, _("affix '%s' is corrupt, possible incorrect count"), 
+                   MsgConv(lang)(achar));
+          return make_err(bad_file_format, affix_file, msg);
+        }
         nptr->achar = achar;
  
         split(dp);
@@ -371,12 +346,12 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
         // key is strip 
         if (dp.key != "0") {
           ParmString s0(iconv(dp.key));
-          max_(max_strip_, s0.size());
-          max_(max_strip_f[(byte)achar], s0.size());
-          nptr->strip = data_buf.dup(s0);
+          MAX(max_strip_, s0.size());
+          MAX(max_strip_f[(byte)achar], s0.size());
+          nptr->strip = strings.dup(s0);
           nptr->stripl = s0.size();
         } else {
-          nptr->strip  = "";
+          nptr->strip= strings.dup("");
           nptr->stripl = 0;
         }
     
@@ -384,46 +359,31 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
         if (dp.key.empty()) goto error;
         // key is affix string or 0 for null
         if (dp.key != "0") {
-          nptr->appnd  = data_buf.dup(iconv(dp.key));
+          nptr->appnd = strings.dup(iconv(dp.key));
           nptr->appndl = strlen(nptr->appnd);
         } else {
-          nptr->appnd  = "";
+          nptr->appnd  = strings.dup("");
           nptr->appndl = 0;
         }
     
         split(dp);
         if (dp.key.empty()) goto error;
         // key is the conditions descriptions
-        char * cond = iconv(dp.key);
-        int cond_len = normalize_cond_str(cond);
-        if (cond_len < 0)
-          return (make_err(invalid_cond, MsgConv(lang)(cond))
-                  .with_file(affix_file, dp.line_num));
-        if (nptr->stripl != 0) {
-          char * cc = cond;
-          if (affix_type == 'S') cc += cond_len - nptr->stripl;
-          if (cond_len < nptr->stripl || 
-              memcmp(cc, nptr->strip, nptr->stripl) != 0)
-            return (make_err(invalid_cond_strip, 
-                             MsgConv(lang)(cond), MsgConv(lang)(nptr->strip))
-                    .with_file(affix_file, dp.line_num));
-        }
-        encodeit(conds_lookup, data_buf, nptr, cond);
+        encodeit(nptr,iconv(dp.key));
     
         // now create SfxEntry or PfxEntry objects and use links to
         // build an ordered (sorted by affix string) list
         if (affix_type == 'P')
-          build_pfxlist(static_cast<PfxEntry *>(nptr));
+          build_pfxlist(static_cast<PfxEntry *>(nptr.release()));
         else
-          build_sfxlist(static_cast<SfxEntry *>(nptr)); 
+          build_sfxlist(static_cast<SfxEntry *>(nptr.release())); 
       }
     }
     continue;
   error:
-    return make_err(corrupt_affix, MsgConv(lang)(achar)).with_file(affix_file, dp.line_num);
-  error_count:
-    return make_err(corrupt_affix, MsgConv(lang)(achar), 
-                    _("Possibly incorrect count.")).with_file(affix_file, dp.line_num);
+    char msg[32];
+    snprintf(msg, 32, _("Affix '%s' is corrupt"), MsgConv(lang)(achar));
+    return make_err(other_error, msg).with_file(affix_file, dp.line_num);
   }
   afflst.close();
 
@@ -452,14 +412,12 @@ PosibErr<void> AffixMgr::parse_file(const char * affpath, Conv & iconv)
   // prefix is not found (call it nextne).
 
   // Since we have built ordered lists, all that remains is to
-  // properly initialize the nextne and nexteq pointers that relate
+  // properly intialize the nextne and nexteq pointers that relate
   // them
 
   process_pfx_order();
   process_sfx_order();
 
-  //CERR.printf("%u\n", data_buf.calc_size()/1024);
-
   return no_err;
 
 }
@@ -492,6 +450,7 @@ PosibErr<void> AffixMgr::build_pfxlist(PfxEntry* pfxptr)
   return no_err;
 }
 
+
 // we want to be able to quickly access suffix information
 // both by suffix flag, and sorted by the reverse of the
 // suffix string itself; so we need to set up two indexes
@@ -500,14 +459,12 @@ PosibErr<void> AffixMgr::build_sfxlist(SfxEntry* sfxptr)
 {
   SfxEntry * ptr;
   SfxEntry * ep = sfxptr;
-  char * tmp = (char *)data_buf.alloc(sfxptr->appndl + 1);
-  sfxptr->rappnd = tmp;
-
+  sfxptr->rappnd = (char *)strings.alloc(sfxptr->appndl + 1);
   // reverse the string
-  char * dest = tmp + sfxptr->appndl;
-  *dest-- = 0;
-  const char * src = sfxptr->appnd;
-  for (; dest >= tmp; --dest, ++src)
+  sfxptr->rappnd[sfxptr->appndl] = 0;
+  for (char * dest = sfxptr->rappnd + sfxptr->appndl - 1, * src = sfxptr->appnd;
+       dest >= sfxptr->rappnd;
+       --dest, ++src)
     *dest = *src;
 
   /* get the right starting point */
@@ -563,7 +520,7 @@ PosibErr<void> AffixMgr::process_pfx_order()
         ptr->next_eq = ptr->next;
     }
 
-    // now clean up by adding smart search termination strings
+    // now clean up by adding smart search termination strings:
     // if you are already a superset of the previous prefix
     // but not a subset of the next, search can end here
     // so set NextNE properly
@@ -635,37 +592,24 @@ PosibErr<void> AffixMgr::process_sfx_order()
   return no_err;
 }
 
+
+
 // takes aff file condition string and creates the
 // conds array - please see the appendix at the end of the
 // file affentry.cxx which describes what is going on here
 // in much more detail
 
-static void encodeit(CondsLookup & l, ObjStack & buf, 
-                     AffEntry * ptr, char * cs)
+void AffixMgr::encodeit(AffEntry * ptr, const char * cs)
 {
   byte c;
   int i, j, k;
-
-  // see if we already have this conds matrix
-
-  CondsLookup::iterator itr = l.find(cs);
-  if (!(itr == l.end())) {
-    ptr->conds = *itr;
-    return;
-  }
-
-  Conds * cds = (Conds *)buf.alloc_bottom(sizeof(Conds));
-  cds->str = buf.dup(cs);
-  l.insert(cds);
-  ptr->conds = cds;
-
   int nc = strlen(cs);
   VARARRAYM(byte, mbr, nc + 1, MAXLNLEN);
 
-  // now clear the conditions array
-  memset(cds->conds, 0, sizeof(cds->conds));
+  // now clear the conditions array */
+  for (i=0;i<SETSIZE;i++) ptr->conds[i] = (byte) 0;
 
-  // now parse the string to create the conds array
+  // now parse the string to create the conds array */
   
   int neg = 0;   // complement indicator
   int grp = 0;   // group indicator
@@ -675,7 +619,7 @@ static void encodeit(CondsLookup & l, ObjStack & buf,
 
   // if no condition just return
   if (strcmp(cs,".")==0) {
-    cds->num = 0;
+    ptr->numconds = 0;
     return;
   }
 
@@ -720,14 +664,14 @@ static void encodeit(CondsLookup & l, ObjStack & buf,
           // set the proper bits in the condition array vals for those chars
           for (j=0;j<nm;j++) {
             k = (unsigned int) mbr[j];
-            cds->conds[k] = cds->conds[k] | (1 << n);
+            ptr->conds[k] = ptr->conds[k] | (1 << n);
           }
         } else {
           // complement so set all of them and then unset indicated ones
-          for (j=0;j<SETSIZE;j++) cds->conds[j] = cds->conds[j] | (1 << n);
+          for (j=0;j<SETSIZE;j++) ptr->conds[j] = ptr->conds[j] | (1 << n);
           for (j=0;j<nm;j++) {
             k = (unsigned int) mbr[j];
-            cds->conds[k] = cds->conds[k] & ~(1 << n);
+            ptr->conds[k] = ptr->conds[k] & ~(1 << n);
           }
         }
         neg = 0;
@@ -738,9 +682,9 @@ static void encodeit(CondsLookup & l, ObjStack & buf,
         // but first handle special case of . inside condition
         if (c == '.') {
           // wild card character so set them all
-          for (j=0;j<SETSIZE;j++) cds->conds[j] = cds->conds[j] | (1 << n);
+          for (j=0;j<SETSIZE;j++) ptr->conds[j] = ptr->conds[j] | (1 << n);
         } else {  
-          cds->conds[(unsigned int)c] = cds->conds[(unsigned int)c] | (1 << n);
+          ptr->conds[(unsigned int) c] = ptr->conds[(unsigned int)c] | (1 << n);
         }
       }
       n++;
@@ -750,21 +694,20 @@ static void encodeit(CondsLookup & l, ObjStack & buf,
 
     i++;
   }
-  cds->num = n;
+  ptr->numconds = n;
   return;
 }
 
 
 // check word for prefixes
 bool AffixMgr::prefix_check (const LookupInfo & linf, ParmString word, 
-                             CheckInfo & ci, GuessInfo * gi, bool cross) const
+                             CheckInfo & ci, GuessInfo * gi) const
 {
-  if (word.empty()) return false;
  
   // first handle the special case of 0 length prefixes
   PfxEntry * pe = pStart[0];
   while (pe) {
-    if (pe->check(linf,this,word,ci,gi)) return true;
+    if (pe->check(linf,word,ci,gi)) return true;
     pe = pe->next;
   }
   
@@ -774,7 +717,7 @@ bool AffixMgr::prefix_check (const LookupInfo & linf, ParmString word,
 
   while (pptr) {
     if (isSubset(pptr->key(),word)) {
-      if (pptr->check(linf,this,word,ci,gi,cross)) return true;
+      if (pptr->check(linf,word,ci,gi)) return true;
       pptr = pptr->next_eq;
     } else {
       pptr = pptr->next_ne;
@@ -790,7 +733,6 @@ bool AffixMgr::suffix_check (const LookupInfo & linf, ParmString word,
                              CheckInfo & ci, GuessInfo * gi,
                              int sfxopts, AffEntry * ppfx) const
 {
-  if (word.empty()) return false;
 
   // first handle the special case of 0 length suffixes
   SfxEntry * se = sStart[0];
@@ -805,7 +747,8 @@ bool AffixMgr::suffix_check (const LookupInfo & linf, ParmString word,
 
   while (sptr) {
     if (isRevSubset(sptr->key(), word + word.size() - 1, word.size())) {
-      if (sptr->check(linf, word, ci, gi, sfxopts, ppfx)) return true;
+      if (sptr->check(linf, word, ci, gi, sfxopts, ppfx))
+        return true;
       sptr = sptr->next_eq;
     } else {
       sptr = sptr->next_ne;
@@ -819,8 +762,6 @@ bool AffixMgr::suffix_check (const LookupInfo & linf, ParmString word,
 bool AffixMgr::affix_check(const LookupInfo & linf, ParmString word, 
                            CheckInfo & ci, GuessInfo * gi) const
 {
-  if (word.empty()) return false;
-
   // Deal With Case in a semi-intelligent manner
   CasePattern cp = lang->LangImpl::case_pattern(word);
   ParmString pword = word;
@@ -844,28 +785,19 @@ bool AffixMgr::affix_check(const LookupInfo & linf, ParmString word,
   if (prefix_check(linf, pword, ci, gi)) return true;
 
   // if still not found check all suffixes
-  if (suffix_check(linf, sword, ci, gi, 0, NULL)) return true;
-
-  // if still not found check again but with the lower case version
-  // which can make a difference if the entire word matches the cond
-  // string
-  if (cp == FirstUpper) {
-    return suffix_check(linf, pword, ci, gi, 0, NULL);
-  } else {
-    return false;
-  }
+  return suffix_check(linf, sword, ci, gi, 0, NULL);
 }
 
-void AffixMgr::munch(ParmString word, GuessInfo * gi, bool cross) const
+void AffixMgr::munch(ParmString word, CheckList * cl) const
 {
   LookupInfo li(0, LookupInfo::AlwaysTrue);
   CheckInfo ci;
-  gi->reset();
+  cl->reset();
   CasePattern cp = lang->LangImpl::case_pattern(word);
   if (cp == AllUpper) return;
   if (cp != FirstUpper)
-    prefix_check(li, word, ci, gi, cross);
-  suffix_check(li, word, ci, gi, 0, NULL);
+    prefix_check(li, word, ci, &cl->gi);
+  suffix_check(li, word, ci, &cl->gi, 0, NULL);
 }
 
 WordAff * AffixMgr::expand(ParmString word, ParmString aff, 
@@ -898,6 +830,7 @@ WordAff * AffixMgr::expand(ParmString word, ParmString aff,
       cur = cur->next;
       cur->word = newword;
       cur->aff = p->allow_cross() ? csuf : empty;
+      break;
     }
   }
 
@@ -914,7 +847,7 @@ WordAff * AffixMgr::expand(ParmString word, ParmString aff,
   for (WordAff * * cur = &head; cur != end; cur = &(*cur)->next) {
     if ((int)(*cur)->word.size - max_strip_ >= limit) continue;
     byte * nsuf = (byte *)buf.alloc(nsuf_s);
-    expand_suffix((*cur)->word, (*cur)->aff, buf, limit, nsuf, &very_end, word);
+    expand_suffix((*cur)->word, (*cur)->aff, buf, limit, nsuf, &very_end);
     (*cur)->aff = nsuf;
   }
 
@@ -923,30 +856,27 @@ WordAff * AffixMgr::expand(ParmString word, ParmString aff,
 
 WordAff * AffixMgr::expand_suffix(ParmString word, const byte * aff, 
                                   ObjStack & buf, int limit,
-                                  byte * new_aff, WordAff * * * l,
-                                  ParmString orig_word) const
+                                  byte * new_aff, WordAff * * * l) const
 {
   WordAff * head = 0;
   if (l) head = **l;
   WordAff * * cur = l ? *l : &head;
-  bool expanded     = false;
-  bool not_expanded = false;
-  if (!orig_word) orig_word = word;
 
   while (*aff) {
-    if ((int)word.size() - max_strip_f[*aff] < limit) {
-      for (SfxEntry * p = sFlag[*aff]; p; p = p->flag_next) {
-        SimpleString newword = p->add(word, buf, limit, orig_word);
-        if (!newword) continue;
-        if (newword == EMPTY) {not_expanded = true; continue;}
-        *cur = (WordAff *)buf.alloc_bottom(sizeof(WordAff));
-        (*cur)->word = newword;
-        (*cur)->aff  = (const byte *)EMPTY;
-        cur = &(*cur)->next;
-        expanded = true;
-      }
+    if ((int)word.size() - max_strip_f[*aff] >= limit) goto not_expanded;
+    for (SfxEntry * p = sFlag[*aff]; p; p = p->flag_next) {
+      SimpleString newword = p->add(word, buf, limit);
+      if (!newword) continue;
+      if (newword == EMPTY) goto not_expanded;
+      *cur = (WordAff *)buf.alloc_bottom(sizeof(WordAff));
+      (*cur)->word = newword;
+      (*cur)->aff  = (const byte *)EMPTY;
+      cur = &(*cur)->next;
+      goto expanded;
     }
-    if (new_aff && (!expanded || not_expanded)) *new_aff++ = *aff;
+  not_expanded:
+    if (new_aff) *new_aff++ = *aff;
+  expanded:
     ++aff;
   }
   *cur = 0;
@@ -955,67 +885,89 @@ WordAff * AffixMgr::expand_suffix(ParmString word, const byte * aff,
   return head;
 }
 
-CheckAffixRes AffixMgr::check_affix(ParmString word, char aff) const
-{
-  CheckAffixRes res = InvalidAffix;
-  
-  for (PfxEntry * p = pFlag[(unsigned char)aff]; p; p = p->flag_next) {
-    res = InapplicableAffix;
-    if (p->applicable(word)) return ValidAffix;
-  }
-
-  for (SfxEntry * p = sFlag[(unsigned char)aff]; p; p = p->flag_next) {
-    if (res == InvalidAffix) res = InapplicableAffix;
-    if (p->applicable(word)) return ValidAffix;
-  }
-
-  return res;
-}
-
-
 
 //////////////////////////////////////////////////////////////////////
 //
 // LookupInfo
 //
 
-int LookupInfo::lookup (ParmString word, const SensitiveCompare * c, 
-                        char achar, 
-                        WordEntry & o, GuessInfo * gi) const
+static void free_word(WordEntry * w)
+{
+  free((void *)w->word);
+}
+
+static void free_aff(WordEntry * w)
+{
+  free((void *)w->aff);
+}
+
+struct LookupBookkepping
+{
+  const char * aff; // the affix here is a list of valid affixes for
+                    // the word
+  bool alloc;
+  LookupBookkepping() : aff(0), alloc(false) {}
+};
+
+static void append_aff(LookupBookkepping & s, WordEntry & o)
+{
+  size_t s0 = strlen(s.aff);
+  size_t s1 = strlen(o.aff);
+  if (s0 == s1 && memcmp(s.aff, o.aff, s0) == 0) return;
+  char * tmp = (char *)malloc(s0 + s1 + 1);
+  // FIXME: avoid adding duplicate flags
+  memcpy(tmp, o.aff, s1);
+  memcpy(tmp + s1, s.aff, s0);
+  tmp[s0 + s1] = '\0';
+  if (s.alloc) free((void *)s.aff);
+  s.aff = tmp;
+  s.alloc = true;
+}
+
+bool LookupInfo::lookup (ParmString word, WordEntry & o) const
 {
   SpellerImpl::WS::const_iterator i = begin;
-  const char * g = 0;
+  const char * w = 0;
+  LookupBookkepping s;
   if (mode == Word) {
     do {
-      (*i)->lookup(word, c, o);
-      for (;!o.at_end(); o.adv()) {
-        if (TESTAFF(o.aff, achar))
-          return 1;
-        else
-          g = o.word;
+      if (i->dict->lookup(word, o, i->compare)) {
+        w = o.word;
+        if (s.aff == 0) s.aff = o.aff;
+        else append_aff(s, o); // this should not be a very common case
       }
       ++i;
     } while (i != end);
   } else if (mode == Clean) {
     do {
-      (*i)->clean_lookup(word, o);
-      for (;!o.at_end(); o.adv()) {
-        if (TESTAFF(o.aff, achar))
-          return 1;
-        else
-          g = o.word;
+      if (i->dict->clean_lookup(word, o)) {
+        w = o.word;
+        if (s.aff == 0) s.aff = o.aff;
+        else append_aff(s, o); // this should not be a very common case
       }
       ++i;
     } while (i != end);
-  } else if (gi) {
-    g = gi->dup(word);
+//   } else if (mode == Soundslike) {
+//     do {
+//       if (i->dict->soundslike_lookup(word, o)) {
+//         w = o.word;
+//         if (s.aff == 0) s.aff = o.aff;
+//         else append_aff(s, o); // this should not be a very common case
+//         while (o.adv()) append_aff(s, o); // neither should this
+//       }
+//       ++i;
+//    } while (i != end);
+  } else {
+    o.word = strdup(word);
+    o.aff  = word + strlen(word);
+    //o.free_ = free_word; //FiXME this isn't right....
+    return true;
   }
-  if (gi && g) {
-    CheckInfo * ci = gi->add();
-    ci->word = g;
-    return -1;
-  }
-  return 0;
+  if (!w) return false;
+  o.word = w;
+  o.aff = s.aff;
+  if (s.alloc) o.free_ = free_aff;
+  return true;
 }
 
 //////////////////////////////////////////////////////////////////////
@@ -1023,33 +975,18 @@ int LookupInfo::lookup (ParmString word, const SensitiveCompare * c,
 // Affix Entry
 //
 
-bool PfxEntry::applicable(SimpleString word) const
-{
-  unsigned int cond;
-  /* make sure all conditions match */
-  if ((word.size > stripl) && (word.size >= conds->num)) {
-    const byte * cp = (const byte *) word.str;
-    for (cond = 0;  cond < conds->num;  cond++) {
-      if ((conds->get(*cp++) & (1 << cond)) == 0)
-        break;
-    }
-    if (cond >= conds->num) return true;
-  }
-  return false;
-}
-
 // add prefix to this word assuming conditions hold
 SimpleString PfxEntry::add(SimpleString word, ObjStack & buf) const
 {
   unsigned int cond;
   /* make sure all conditions match */
-  if ((word.size > stripl) && (word.size >= conds->num)) {
+  if ((word.size > stripl) && (word.size >= numconds)) {
     const byte * cp = (const byte *) word.str;
-    for (cond = 0;  cond < conds->num;  cond++) {
-      if ((conds->get(*cp++) & (1 << cond)) == 0)
+    for (cond = 0;  cond < numconds;  cond++) {
+      if ((conds[*cp++] & (1 << cond)) == 0)
         break;
     }
-    if (cond >= conds->num) {
+    if (cond >= numconds) {
       /* */
       int alen = word.size - stripl;
       char * newword = (char *)buf.alloc(alen + appndl + 1);
@@ -1062,15 +999,14 @@ SimpleString PfxEntry::add(SimpleString word, ObjStack & buf) const
 }
 
 // check if this prefix entry matches 
-bool PfxEntry::check(const LookupInfo & linf, const AffixMgr * pmyMgr,
-                     ParmString word,
-                     CheckInfo & ci, GuessInfo * gi, bool cross) const
+bool PfxEntry::check(const LookupInfo & linf, ParmString word,
+                     CheckInfo & ci, GuessInfo * gi) const
 {
   unsigned int		cond;	// condition number being examined
-  unsigned              tmpl;   // length of tmpword
+  int	                tmpl;   // length of tmpword
   WordEntry             wordinfo;     // hash entry of root word or NULL
   byte *	cp;		
-  VARARRAYM(char, tmpword, word.size()+stripl+1, MAXWORDLEN+1);
+  VARARRAYM(char, tmpword, word.size()+1, MAXWORDLEN+1);
 
   // on entry prefix is 0 length or already matches the beginning of the word.
   // So if the remaining root word has positive length
@@ -1079,7 +1015,7 @@ bool PfxEntry::check(const LookupInfo & linf, const AffixMgr * pmyMgr,
 
   tmpl = word.size() - appndl;
 
-  if ((tmpl > 0) &&  (tmpl + stripl >= conds->num)) {
+  if ((tmpl > 0) &&  (tmpl + stripl >= numconds)) {
 
     // generate new root word by removing prefix and adding
     // back any characters that would have been stripped
@@ -1093,106 +1029,72 @@ bool PfxEntry::check(const LookupInfo & linf, const AffixMgr * pmyMgr,
     // tested
 
     cp = (byte *)tmpword;
-    for (cond = 0;  cond < conds->num;  cond++) {
-      if ((conds->get(*cp++) & (1 << cond)) == 0) break;
+    for (cond = 0;  cond < numconds;  cond++) {
+      if ((conds[*cp++] & (1 << cond)) == 0) break;
     }
 
     // if all conditions are met then check if resulting
     // root word in the dictionary
 
-    if (cond >= conds->num) {
+    if (cond >= numconds) {
       CheckInfo * lci = 0;
-      CheckInfo * guess = 0;
       tmpl += stripl;
+      if (linf.lookup(tmpword, wordinfo)) {
 
-      int res = linf.lookup(tmpword, &linf.sp->s_cmp_end, achar, wordinfo, gi);
-
-      if (res == 1) {
-
-        lci = &ci;
-        lci->word = wordinfo.word;
-        goto quit;
-        
-      } else if (res == -1) {
-
-        guess = gi->head;
-
-      }
-      
-      // prefix matched but no root word was found 
-      // if XPRODUCT is allowed, try again but now 
-      // cross checked combined with a suffix
-      
-      if (gi)
-        lci = gi->head;
-      
-      if (cross && xpflg & XPRODUCT) {
-        if (pmyMgr->suffix_check(linf, ParmString(tmpword, tmpl), 
-                                 ci, gi,
-                                 XPRODUCT, (AffEntry *)this)) {
+        if (TESTAFF(wordinfo.aff, achar))
           lci = &ci;
-          
-        } else if (gi) {
-          
-          CheckInfo * stop = lci;
-          for (lci = gi->head; 
-               lci != stop; 
-               lci = const_cast<CheckInfo *>(lci->next)) 
-          {
-            lci->pre_flag = achar;
-            lci->pre_strip_len = stripl;
-            lci->pre_add_len = appndl;
-            lci->pre_add = appnd;
+        else if (gi)
+          lci = gi->add();
+
+        if (lci)
+          lci->word = wordinfo.word;
+
+      } else {
+              
+        // prefix matched but no root word was found 
+        // if XPRODUCT is allowed, try again but now 
+        // cross checked combined with a suffix
+                
+        if (gi)
+          lci = gi->last;
+                
+        if (xpflg & XPRODUCT) {
+          if (pmyMgr->suffix_check(linf, ParmString(tmpword, tmpl), 
+                                   ci, gi,
+                                   XPRODUCT, (AffEntry *)this))
+            lci = &ci;
+          else if (gi && gi->last != lci) {
+            while (lci = const_cast<CheckInfo *>(lci->next), lci) {
+              lci->pre_flag = achar;
+              lci->pre_strip_len = stripl;
+              lci->pre_add_len = appndl;
+              lci->pre_add = appnd;
+            }
           }
-          
-        } else {
-          
-          lci = 0;
-          
         }
       }
-    
-      if (guess)
-        lci = guess;
-      
-    quit:
+              
       if (lci) {
         lci->pre_flag = achar;
         lci->pre_strip_len = stripl;
         lci->pre_add_len = appndl;
         lci->pre_add = appnd;
       }
-      if (lci == &ci) return true;
+      if (lci ==&ci) return true;
     }
   }
   return false;
 }
 
-bool SfxEntry::applicable(SimpleString word) const
-{
-  int cond;
-  /* make sure all conditions match */
-  if ((word.size > stripl) && (word.size >= conds->num)) {
-    const byte * cp = (const byte *) (word + word.size);
-    for (cond = conds->num; --cond >=0; ) {
-      if ((conds->get(*--cp) & (1 << cond)) == 0)
-        break;
-    }
-    if (cond < 0) return true;
-  }
-  return false;
-}
-
 // add suffix to this word assuming conditions hold
-SimpleString SfxEntry::add(SimpleString word, ObjStack & buf, 
-                           int limit, SimpleString orig_word) const
+SimpleString SfxEntry::add(SimpleString word, ObjStack & buf, int limit) const
 {
   int cond;
   /* make sure all conditions match */
-  if ((orig_word.size > stripl) && (orig_word.size >= conds->num)) {
-    const byte * cp = (const byte *) (orig_word + orig_word.size);
-    for (cond = conds->num; --cond >=0; ) {
-      if ((conds->get(*--cp) & (1 << cond)) == 0)
+  if ((word.size > stripl) && (word.size >= numconds)) {
+    const byte * cp = (const byte *) (word + word.size);
+    for (cond = numconds; --cond >=0; ) {
+      if ((conds[*--cp] & (1 << cond)) == 0)
         break;
     }
     if (cond < 0) {
@@ -1213,13 +1115,14 @@ bool SfxEntry::check(const LookupInfo & linf, ParmString word,
                      CheckInfo & ci, GuessInfo * gi,
                      int optflags, AffEntry* ppfx)
 {
-  unsigned              tmpl;		 // length of tmpword 
+  int	                tmpl;		 // length of tmpword 
   int			cond;		 // condition beng examined
   WordEntry             wordinfo;        // hash entry pointer
   byte *	cp;
-  VARARRAYM(char, tmpword, word.size()+stripl+1, MAXWORDLEN+1);
+  VARARRAYM(char, tmpword, word.size()+1, MAXWORDLEN+1);
   PfxEntry* ep = (PfxEntry *) ppfx;
 
+
   // if this suffix is being cross checked with a prefix
   // but it does not support cross products skip it
 
@@ -1233,7 +1136,7 @@ bool SfxEntry::check(const LookupInfo & linf, ParmString word,
 
   tmpl = word.size() - appndl;
 
-  if ((tmpl > 0)  &&  (tmpl + stripl >= conds->num)) {
+  if ((tmpl > 0)  &&  (tmpl + stripl >= numconds)) {
 
     // generate new root word by removing suffix and adding
     // back any characters that would have been stripped or
@@ -1252,8 +1155,8 @@ bool SfxEntry::check(const LookupInfo & linf, ParmString word,
     // this file for more info on exactly what is being
     // tested
 
-    for (cond = conds->num;  --cond >= 0; ) {
-      if ((conds->get(*--cp) & (1 << cond)) == 0) break;
+    for (cond = numconds;  --cond >= 0; ) {
+      if ((conds[*--cp] & (1 << cond)) == 0) break;
     }
 
     // if all conditions are met then check if resulting
@@ -1262,29 +1165,24 @@ bool SfxEntry::check(const LookupInfo & linf, ParmString word,
     if (cond < 0) {
       CheckInfo * lci = 0;
       tmpl += stripl;
-      const SensitiveCompare * cmp = 
-        optflags & XPRODUCT ? &linf.sp->s_cmp_middle : &linf.sp->s_cmp_begin;
-      int res = linf.lookup(tmpword, cmp, achar, wordinfo, gi);
-      if (res == 1
-          && ((optflags & XPRODUCT) == 0 || TESTAFF(wordinfo.aff, ep->achar)))
-      {
-        lci = &ci;
-        lci->word = wordinfo.word;
-      } else if (res == 1 && gi) {
-        lci = gi->add();
-        lci->word = wordinfo.word;
-      } else if (res == -1) { // gi must be defined
-        lci = gi->head;
-      }
+      if (linf.lookup(tmpword, wordinfo)) {
+        if (TESTAFF(wordinfo.aff, achar) && 
+            ((optflags & XPRODUCT) == 0 || 
+             TESTAFF(wordinfo.aff, ep->achar)))
+          lci = &ci;
+        else if (gi)
+          lci = gi->add();
 
-      if (lci) {
-        lci->suf_flag = achar;
-        lci->suf_strip_len = stripl;
-        lci->suf_add_len = appndl;
-        lci->suf_add = appnd;
+        if (lci) {
+          lci->word = wordinfo.word;
+          lci->suf_flag = achar;
+          lci->suf_strip_len = stripl;
+          lci->suf_add_len = appndl;
+          lci->suf_add = appnd;
+        }
+        
+        if (lci == &ci) return true;
       }
-      
-      if (lci == &ci) return true;
     }
   }
   return false;
@@ -1408,7 +1306,7 @@ specific character position in the word.
 For prefixes, it does this by setting bit 0 if that char is valid 
 in the first position, bit 1 if valid in the second position, and so on. 
 
-If a bit is not set, then that char is not valid for that position in the
+If a bit is not set, then that char is not valid for that postion in the
 word.
 
 If working with suffixes bit 0 is used for the character closest 
