@@ -1,9 +1,8 @@
-// Copyright 2000,2011 by Kevin Atkinson under the terms of the LGPL
+// Copyright 2000 by Kevin Atkinson under the terms of the LGPL
 
 #ifndef ASPELLER_DATA__HPP
 #define ASPELLER_DATA__HPP
 
-#include "ndebug.hpp"
 #include <assert.h>
 
 #include "copy_ptr.hpp"
@@ -27,10 +26,11 @@ namespace acommon {
 
 namespace aspeller {
 
-  class Dictionary;
-  class DictList;
+  struct LocalDict;
+  struct LocalDictInfo;
+  class LocalDictList;
   typedef Enumeration<WordEntry *> WordEntryEnumeration;
-  typedef Enumeration<Dictionary *> DictsEnumeration;
+  typedef Enumeration<const LocalDict *> DictsEnumeration;
 
   class SoundslikeEnumeration 
   {
@@ -66,7 +66,7 @@ namespace aspeller {
     class Id;
   protected:
     CopyPtr<Id> id_;
-    virtual void set_lang_hook(Config &) {}
+    virtual void set_lang_hook(const Config *) {}
     
   public:
     typedef Id CacheKey;
@@ -83,7 +83,7 @@ namespace aspeller {
 
     const Id & id() {return *id_;}
     PosibErr<void> check_lang(ParmString lang);
-    PosibErr<void> set_check_lang(ParmString lang, Config &);
+    PosibErr<void> set_check_lang(ParmString lang, const Config *);
     const LangImpl * lang() const {return lang_;};
     const Language * language() const {return lang_;};
     const char * lang_name() const;
@@ -98,8 +98,8 @@ namespace aspeller {
 
     const char * file_name() const {return file_name_.path.c_str();}
     // returns any additional dictionaries that are also used
-    virtual PosibErr<void> load(ParmString, Config &, DictList * = 0, 
-                                SpellerImpl * = 0);
+    virtual PosibErr<void> load(ParmString, const Config &, LocalDictList * = 0, 
+                                SpellerImpl * = 0, const LocalDictInfo * = 0);
 
     virtual PosibErr<void> merge(ParmString);
     virtual PosibErr<void> synchronize();
@@ -108,12 +108,7 @@ namespace aspeller {
     virtual PosibErr<void> clear();
 
     bool affix_compressed;
-    bool invisible_soundslike; // true when words are grouped by the
-                               // soundslike but soundslike data is not
-                               // actually stored with the word
-    bool soundslike_root_only; // true when affix compression is used AND
-                               // the stored soundslike corresponds to the
-                               // root word only
+    bool have_soundslike; // only true when there is true phonet data
     bool fast_scan;  // can effectly scan for all soundslikes (or
                      // clean words if have_soundslike is false)
                      // with an edit distance of 1 or 2
@@ -122,7 +117,7 @@ namespace aspeller {
     
     typedef WordEntryEnumeration        Enum;
     typedef const char *                Value;
-    typedef unsigned int                Size;
+    typedef size_t                      Size;
 
     StringEnumeration * elements() const;
 
@@ -130,8 +125,8 @@ namespace aspeller {
     virtual Size   size()     const;
     virtual bool   empty()    const;
   
-    virtual bool lookup (ParmString word, const SensitiveCompare *, 
-                         WordEntry &) const;
+    virtual bool lookup (ParmString word, WordEntry &,
+                         const SensitiveCompare &) const;
     
     virtual bool clean_lookup(ParmString, WordEntry &) const;
 
@@ -171,21 +166,38 @@ namespace aspeller {
     return !(rhs == lhs);
   }
 
-  class DictList {
+  struct LocalDictInfo 
+  {
+    SensitiveCompare compare;
+    ConvertWord      convert;
+    void set_language(const Language * l);
+    void set(const Language * l, const Config & c, bool strip = false);
+    void set(const LocalDictInfo & li) {operator=(li);}
+  };
+
+  struct LocalDict : public LocalDictInfo {
+    Dict * dict;
+    LocalDict(Dict * d = 0) : dict(d) {}
+    LocalDict(Dict * d, LocalDictInfo li)
+      : LocalDictInfo(li), dict(d) {}
+    operator bool () const {return dict != 0;}
+  };
+
+  class LocalDictList {
     // well a stack at the moment but it may eventually become a list
     // NOT necessarily first in first out
-    Vector<Dict *> data;
+    Vector<LocalDict> data;
   private:
-    DictList(const DictList &);
-    void operator= (const DictList &);
+    LocalDictList(const LocalDictList &);
+    void operator= (const LocalDictList &);
   public:
     // WILL take ownership of the dict
-    DictList() {}
-    void add(Dict * o) {data.push_back(o);}
-    Dict * last() {return data.back();}
+    LocalDictList() {}
+    void add(const LocalDict & o) {data.push_back(o);}
+    LocalDict & last() {return data.back();}
     void pop() {data.pop_back();}
     bool empty() {return data.empty();}
-    ~DictList() {for (; !empty(); pop()) last()->release();}
+    ~LocalDictList() {for (; !empty(); pop()) last().dict->release();}
   };
 
   typedef unsigned int DataType;
@@ -195,13 +207,16 @@ namespace aspeller {
   static const DataType DT_Multi        = 1<<3;
   static const DataType DT_Any          = 0xFF;
 
+  // stores result in LocalDict
   // any new extra dictionaries that were loaded will be ii
-  PosibErr<Dict *> add_data_set(ParmString file_name,
-                                Config &,
-                                DictList * other_dicts = 0,
-                                SpellerImpl * = 0,
-                                ParmString dir = 0,
-                                DataType allowed = DT_Any);
+  PosibErr<void> add_data_set(ParmString file_name,
+                              const Config &,
+                              LocalDict &,
+                              LocalDictList * other_dicts = 0,
+                              SpellerImpl * = 0,
+                              const LocalDictInfo * local_info = 0,
+                              ParmString dir = 0,
+                              DataType allowed = DT_Any);
   
   // implemented in readonly_ws.cc
   Dictionary * new_default_readonly_dict();
