@@ -1,9 +1,3 @@
-// This file is part of The New Aspell
-// Copyright (C) 2000,2011 by Kevin Atkinson under the GNU LGPL
-// license version 2.0 or 2.1.  You should have received a copy of the
-// LGPL license along with this library if you did not you can find it
-// at http://www.gnu.org/.
-
 #include <time.h>
 
 #include "hash-t.hpp"
@@ -27,120 +21,39 @@ using namespace std;
 using namespace aspeller;
 using namespace acommon;
 
-typedef const char * Str;
-typedef unsigned char byte;
-
-struct Hash {
-  InsensitiveHash<> f;
-  Hash(const Language * l) : f(l) {}
-  size_t operator() (Str s) const {
-    return f(s);
-  }
-};
-
-struct Equal {
-  InsensitiveEqual f;
-  Equal(const Language * l) : f(l) {}
-  bool operator() (Str a, Str b) const {
-    return f(a, b);
-  }
-};
-
-void write_n_escape(FStream & o, const char * str) {
-  while (*str != '\0') {
-    if (*str == '\n') o << "\\n";
-    else if (*str == '\r') o << "\\r";
-    else if (*str == '\\') o << "\\\\";
-    else o << *str;
-    ++str;
-  }
-}
-
-static inline char f_getc(FStream & in) {
-  int c = in.get();
-  return c == EOF ? '\0' : (char)c;
-}
-  
-bool getline_n_unescape(FStream & in, String & str, char delem) {
-  str.clear();
-  char c = f_getc(in);
-  if (!c) return false;
-  while (c && c != delem) {
-    if (c == '\\') {
-      c = f_getc(in);
-      if (c == 'n') str.append('\n');
-      else if (c == 'r') str.append('\r');
-      else if (c == '\\') str.append('\\');
-      else {str.append('\\'); continue;}
-    } else {
-      str.append(c);
-    }
-    c = f_getc(in);
-  }
-  return true;
-}
-
-bool getline_n_unescape(FStream & in, DataPair & d, String & buf)
-{
-  if (!getline_n_unescape(in, buf, '\n')) return false;
-  d.value.str  = buf.mstr();
-  d.value.size = buf.size();
-  return true;
-}
-
-typedef Vector<Str> StrVector;
-
-typedef hash_multiset<Str,Hash,Equal> WordLookup;
-typedef hash_map<Str,StrVector>  SoundslikeLookup;
-
 class WritableBase : public Dictionary {
 protected:
   String suffix;
   String compatibility_suffix;
     
   time_t cur_file_date;
-  
+    
   String compatibility_file_name;
     
-  WritableBase(BasicType t, const char * n, const char * s, const char * cs, const Config & cfg)
+  WritableBase(BasicType t, const char * n, const char * s, const char * cs)
     : Dictionary(t,n),
-      suffix(s), compatibility_suffix(cs),
-      use_soundslike(true) {
-    fast_lookup = true;
-    validate_words = cfg.retrieve_bool("validate-words");
-  }
+      suffix(s), compatibility_suffix(cs) {}
   virtual ~WritableBase() {}
-  
+    
   virtual PosibErr<void> save(FStream &, ParmString) = 0;
-  virtual PosibErr<void> merge(FStream &, ParmString, Config * = 0) = 0;
+  virtual PosibErr<void> merge(FStream &, ParmString, const Config * = 0) = 0;
     
   PosibErr<void> save2(FStream &, ParmString);
   PosibErr<void> update(FStream &, ParmString);
   PosibErr<void> save(bool do_update);
   PosibErr<void> update_file_date_info(FStream &);
-  PosibErr<void> load(ParmString, Config &, DictList *, SpellerImpl *);
+  PosibErr<void> load(ParmString, const Config &, LocalDictList *,
+                      SpellerImpl *, const LocalDictInfo *);
   PosibErr<void> merge(ParmString);
   PosibErr<void> save_as(ParmString);
-  PosibErr<void> clear();
 
   String file_encoding;
   ConvObj iconv;
   ConvObj oconv;
-  PosibErr<void> set_file_encoding(ParmString, Config & c);
+  PosibErr<void> set_file_encoding(ParmString, const Config * c);
 
   PosibErr<void> synchronize() {return save(true);}
   PosibErr<void> save_noupdate() {return save(false);}
-
-  bool use_soundslike;
-  StackPtr<WordLookup> word_lookup;
-  SoundslikeLookup     soundslike_lookup_;
-  ObjStack             buffer;
- 
-  void set_lang_hook(Config & c) {
-    set_file_encoding(lang()->data_encoding(), c);
-    word_lookup.reset(new WordLookup(10, Hash(lang()), Equal(lang())));
-    use_soundslike = lang()->have_soundslike();
-  }
 };
 
 PosibErr<void> WritableBase::update_file_date_info(FStream & f) {
@@ -149,8 +62,9 @@ PosibErr<void> WritableBase::update_file_date_info(FStream & f) {
   return no_err;
 }
   
-PosibErr<void> WritableBase::load(ParmString f0, Config & config,
-                                  DictList *, SpellerImpl *)
+PosibErr<void> WritableBase::load(ParmString f0, const Config & config,
+                                  LocalDictList *,
+                                  SpellerImpl *, const LocalDictInfo *)
 {
   set_file_name(f0);
   const String f = file_name();
@@ -163,8 +77,7 @@ PosibErr<void> WritableBase::load(ParmString f0, Config & config,
     // ^^ FIXME 
     RET_ON_ERR(merge(in, f, &config));
       
-  } else if (f.size() >= suffix.size() &&
-             f.substr(f.size()-suffix.size(),suffix.size())
+  } else if (f.substr(f.size()-suffix.size(),suffix.size()) 
              == suffix) {
       
     compatibility_file_name = f.substr(0,f.size() - suffix.size());
@@ -190,8 +103,8 @@ PosibErr<void> WritableBase::load(ParmString f0, Config & config,
 PosibErr<void> WritableBase::merge(ParmString f0) {
   FStream in;
   Dict::FileName fn(f0);
-  RET_ON_ERR(open_file_readlock(in, fn.path()));
-  RET_ON_ERR(merge(in, fn.path()));
+  RET_ON_ERR(open_file_readlock(in, fn.path));
+  RET_ON_ERR(merge(in, fn.path));
   return no_err;
 }
 
@@ -248,19 +161,12 @@ PosibErr<void> WritableBase::save(bool do_update) {
   return no_err;
 }
 
-PosibErr<void> WritableBase::clear() {
-  word_lookup->clear();
-  soundslike_lookup_.clear();
-  buffer.reset();
-  return no_err;
-}
-
-PosibErr<void> WritableBase::set_file_encoding(ParmString enc, Config & c)
+PosibErr<void> WritableBase::set_file_encoding(ParmString enc, const Config * c)
 {
   if (enc == file_encoding) return no_err;
-  if (enc == "") enc = lang()->charmap();
-  RET_ON_ERR(iconv.setup(c, enc, lang()->charmap(), NormFrom));
-  RET_ON_ERR(oconv.setup(c, lang()->charmap(), enc, NormTo));
+  if (enc == "") enc = lang()->charset();
+  RET_ON_ERR(iconv.setup(*c, enc, lang()->charset()));
+  RET_ON_ERR(oconv.setup(*c, lang()->charset(), enc));
   if (iconv || oconv) 
     file_encoding = enc;
   else
@@ -274,49 +180,52 @@ PosibErr<void> WritableBase::set_file_encoding(ParmString enc, Config & c)
 //  Common Stuff
 //
 
-// a word is stored in memory as follows
-//   <word info><size><word...><null>
-// the hash table points to the word and not the start of the block
+typedef const char * Str;
 
-static inline void set_word(WordEntry & res, Str w)
-{
-  res.word      = w;
-  res.word_size = (byte)w[-1];
-  res.word_info = (byte)w[-2];
-  res.aff       = "";
-}
+struct Hash {
+  InsensitiveHash f;
+  Hash(const Language * l) : f(l) {}
+  size_t operator() (Str s) const {
+    return f(s);
+  }
+};
 
-// a soundslike is stored in memory as follows
-//   <word info><size><sl...><null>
-// the hash table points to the sl and not the start of the block
+struct Equal {
+  InsensitiveEqual f;
+  Equal(const Language * l) : f(l) {}
+  bool operator() (Str a, Str b) const
+  {
+    return f(a, b);
+  }
+};
 
-static inline void set_sl(WordEntry & res, Str w)
-{
-  res.word      = w;
-  res.word_size = (byte)w[-1];
-}
+typedef Vector<Str> StrVector;
+
+typedef hash_set<Str,Hash,Equal> WordLookup;
+typedef hash_map<Str,StrVector>  SoundslikeLookup;
 
 static void soundslike_next(WordEntry * w)
 {
-  const Str * & i   = (const Str * &)(w->intr[0]);
-  const Str *   end = (const Str *  )(w->intr[1]);
-  set_word(*w, *i);
+  const char * const * i   = (const char * const *)(w->intr[0]);
+  const char * const * end = (const char * const *)(w->intr[1]);
+  w->word = *i;
   ++i;
   if (i == end) w->adv_ = 0;
 }
 
 static void sl_init(const StrVector * tmp, WordEntry & o)
 {
-  const Str * i   = tmp->pbegin();
-  const Str * end = tmp->pend();
-  set_word(o, *i);
+  const char * const * i   = tmp->pbegin();
+  const char * const * end = tmp->pend();
+  o.word = *i;
+  o.aff  = "";
   ++i;
   if (i != end) {
     o.intr[0] = (void *)i;
     o.intr[1] = (void *)end;
     o.adv_ = soundslike_next;
   } else {
-    o.intr[0] = 0;
+     o.intr[0] = 0;
   }
 }
 
@@ -335,7 +244,7 @@ struct SoundslikeElements : public SoundslikeEnumeration {
 
   WordEntry * next(int) {
     if (i == end) return 0;
-    set_sl(d, i->first);
+    d.word = i->first;
     d.intr[0] = (void *)(&i->second);
     ++i;
     return &d;
@@ -353,11 +262,12 @@ struct CleanElements : public SoundslikeEnumeration {
 
   CleanElements(Itr i0, Itr end0) : i(i0), end(end0) {
     d.what = WordEntry::Word;
+    d.aff  = "";
   }
 
   WordEntry * next(int) {
     if (i == end) return 0;
-    set_word(d, *i);
+    d.word = *i;
     ++i;
     return &d;
   }
@@ -370,7 +280,7 @@ struct ElementsParms {
   WordEntry data;
   ElementsParms(Iterator e) : end_(e) {}
   bool endf(Iterator i) const {return i==end_;}
-  Value deref(Iterator i) {set_word(data, *i); return &data;}
+  Value deref(Iterator i) {data.word = *i; return &data;}
   static Value end_state() {return 0;}
 };
 
@@ -382,23 +292,34 @@ struct ElementsParms {
 class WritableDict : public WritableBase
 {
 public: //but don't use
-  PosibErr<void> save(FStream &, ParmString);
-  PosibErr<void> merge(FStream &, ParmString, Config * config);
+  StackPtr<WordLookup> word_lookup;
+  SoundslikeLookup     soundslike_lookup_;
+  ObjStack             buffer;
 
+  PosibErr<void> save(FStream &, ParmString);
+  PosibErr<void> merge(FStream &, ParmString, const Config * config);
+
+protected:
+  void set_lang_hook(const Config * c) {
+    set_file_encoding(lang()->data_encoding(), c);
+    word_lookup.reset(new WordLookup(10, Hash(lang()), Equal(lang())));
+  }
+    
 public:
 
-  WritableDict(const Config & cfg)
-    : WritableBase(basic_dict, "WritableDict", ".pws", ".per", cfg) {}
+  WritableDict() : WritableBase(basic_dict, "WritableDict", ".pws", ".per") {
+    have_soundslike = true; fast_lookup = true;
+  }
 
   Size   size()     const;
   bool   empty()    const;
   
-  PosibErr<void> add(ParmString w) {return Dictionary::add(w);}
+  using Dictionary::add;
   PosibErr<void> add(ParmString w, ParmString s);
 
-  bool lookup(ParmString word, const SensitiveCompare *, WordEntry &) const;
+  bool lookup (ParmString word, WordEntry &, const SensitiveCompare &) const;
 
-  bool clean_lookup(ParmString sondslike, WordEntry &) const;
+  bool clean_lookup(const char * sondslike, WordEntry &) const;
 
   bool soundslike_lookup(const WordEntry & soundslike, WordEntry &) const;
   bool soundslike_lookup(ParmString soundslike, WordEntry &) const;
@@ -418,15 +339,16 @@ bool WritableDict::empty() const
   return word_lookup->empty();
 }
 
-bool WritableDict::lookup(ParmString word, const SensitiveCompare * c,
-                          WordEntry & o) const
+bool WritableDict::lookup(ParmString word, WordEntry & o,
+                        const SensitiveCompare & c) const
 {
   o.clear();
   pair<WordLookup::iterator, WordLookup::iterator> p(word_lookup->equal_range(word));
   while (p.first != p.second) {
-    if ((*c)(word,*p.first)) {
+    if (c(word,*p.first)) {
       o.what = WordEntry::Word;
-      set_word(o, *p.first);
+      o.word = *p.first;
+      o.aff  = "";
       return true;
     }
     ++p.first;
@@ -434,20 +356,21 @@ bool WritableDict::lookup(ParmString word, const SensitiveCompare * c,
   return false;
 }
 
-bool WritableDict::clean_lookup(ParmString sl, WordEntry & o) const
+bool WritableDict::clean_lookup(const char * sl, WordEntry & o) const
 {
   o.clear();
   pair<WordLookup::iterator, WordLookup::iterator> p(word_lookup->equal_range(sl));
   if (p.first == p.second) return false;
   o.what = WordEntry::Word;
-  set_word(o, *p.first);
+  o.word = *p.first;
+  o.aff  = "";
   return true;
   // FIXME: Deal with multiple entries
 }  
 
 bool WritableDict::soundslike_lookup(const WordEntry & word, WordEntry & o) const 
 {
-  if (use_soundslike) {
+  if (have_soundslike) {
 
     const StrVector * tmp 
       = (const StrVector *)(word.intr[0]);
@@ -460,8 +383,6 @@ bool WritableDict::soundslike_lookup(const WordEntry & word, WordEntry & o) cons
       
     o.what = WordEntry::Word;
     o.word = word.word;
-    o.word_size = word.word_size;
-    o.word_info = word.word_info;
     o.aff  = "";
     
   }
@@ -470,7 +391,7 @@ bool WritableDict::soundslike_lookup(const WordEntry & word, WordEntry & o) cons
 
 bool WritableDict::soundslike_lookup(ParmString word, WordEntry & o) const 
 {
-  if (use_soundslike) {
+  if (have_soundslike) {
 
     o.clear();
     SoundslikeLookup::const_iterator i = soundslike_lookup_.find(word);
@@ -490,12 +411,12 @@ bool WritableDict::soundslike_lookup(ParmString word, WordEntry & o) const
 }
 
 SoundslikeEnumeration * WritableDict::soundslike_elements() const {
-  if (use_soundslike)
+  if (have_soundslike)
     return new SoundslikeElements(soundslike_lookup_.begin(), 
                                   soundslike_lookup_.end());
   else
     return new CleanElements(word_lookup->begin(),
-                             word_lookup->end());
+                                word_lookup->end());
 }
 
 WritableDict::Enum * WritableDict::detailed_elements() const {
@@ -504,32 +425,20 @@ WritableDict::Enum * WritableDict::detailed_elements() const {
 }
 
 PosibErr<void> WritableDict::add(ParmString w, ParmString s) {
-  if (validate_words)
-    RET_ON_ERR(check_if_valid(*lang(),w));
-  else
-    RET_ON_ERR(check_if_sane(*lang(),w));
+  RET_ON_ERR(check_if_valid(*lang(),w));
   SensitiveCompare c(lang());
   WordEntry we;
-  if (WritableDict::lookup(w,&c,we)) return no_err;
-  byte * w2;
-  w2 = (byte *)buffer.alloc(w.size() + 3);
-  *w2++ = lang()->get_word_info(w);
-  *w2++ = w.size();
-  memcpy(w2, w.str(), w.size() + 1);
-  word_lookup->insert((char *)w2);
-  if (use_soundslike) {
-    byte * s2;
-    s2 = (byte *)buffer.alloc(s.size() + 2);
-    *s2++ = s.size();
-    memcpy(s2, s.str(), s.size() + 1);
-    soundslike_lookup_[(char *)s2].push_back((char *)w2);
-  }
+  if (WritableDict::lookup(w,we,c)) return no_err;
+  const char * w2 = buffer.dup(w);
+  word_lookup->insert(w2);
+  if (have_soundslike)
+    soundslike_lookup_[buffer.dup(s)].push_back(w2);
   return no_err;
 }
 
 PosibErr<void> WritableDict::merge(FStream & in, 
-                                   ParmString file_name, 
-                                   Config * config)
+                                 ParmString file_name, 
+                                 const Config * config)
 {
   typedef PosibErr<void> Ret;
   unsigned int ver;
@@ -550,7 +459,7 @@ PosibErr<void> WritableDict::merge(FStream & in,
 
   split(dp);
   {
-    Ret pe = set_check_lang(dp.key, *config);
+    Ret pe = set_check_lang(dp.key, config);
     if (pe.has_err())
       return pe.with_file(file_name);
   }
@@ -559,12 +468,12 @@ PosibErr<void> WritableDict::merge(FStream & in,
 
   split(dp);
   if (dp.key.size > 0)
-    set_file_encoding(dp.key, *config);
+    set_file_encoding(dp.key, config);
   else
-    set_file_encoding("", *config);
+    set_file_encoding("", config);
   
   ConvP conv(iconv);
-  while (getline_n_unescape(in, dp, buf)) {
+  while (getline(in, dp, buf)) {
     if (ver == 10)
       split(dp);
     else
@@ -583,13 +492,16 @@ PosibErr<void> WritableDict::save(FStream & out, ParmString file_name)
   out.printf("personal_ws-1.1 %s %i %s\n", 
              lang_name(), word_lookup->size(), file_encoding.c_str());
 
-  WordLookup::const_iterator i = word_lookup->begin();
-  WordLookup::const_iterator e = word_lookup->end();
+  SoundslikeLookup::const_iterator i = soundslike_lookup_.begin();
+  SoundslikeLookup::const_iterator e = soundslike_lookup_.end();
     
+  StrVector::const_iterator j;
+  
   ConvP conv(oconv);
   for (;i != e; ++i) {
-    write_n_escape(out, conv(*i));
-    out << '\n';
+    for (j = i->second.begin(); j != i->second.end(); ++j) {
+      out.printf("%s\n", conv(*j));
+    }
   }
   return no_err;
 }
@@ -601,18 +513,29 @@ PosibErr<void> WritableDict::save(FStream & out, ParmString file_name)
 
 static inline StrVector * get_vector(Str s) 
 {
-  return (StrVector *)(s - sizeof(StrVector) - 2);
+  return (StrVector *)(s - sizeof(StrVector));
 }
 
 class WritableReplDict : public WritableBase
 {
+private:
+  StackPtr<WordLookup> word_lookup;
+  SoundslikeLookup         soundslike_lookup_;
+  StringBuffer             buffer;
+
   WritableReplDict(const WritableReplDict&);
   WritableReplDict& operator=(const WritableReplDict&);
 
+protected:
+  void set_lang_hook(const Config * c) {
+    set_file_encoding(lang()->data_encoding(), c);
+    word_lookup.reset(new WordLookup(10, Hash(lang()), Equal(lang())));
+  }
+
 public:
-  WritableReplDict(const Config & cfg)
-    : WritableBase(replacement_dict, "WritableReplDict", ".prepl",".rpl", cfg) 
+  WritableReplDict() : WritableBase(replacement_dict, "WritableReplDict", ".prepl",".rpl") 
   {
+    have_soundslike = true;
     fast_lookup = true;
   }
   ~WritableReplDict();
@@ -620,7 +543,7 @@ public:
   Size   size()     const;
   bool   empty()    const;
 
-  bool lookup(ParmString, const SensitiveCompare *, WordEntry &) const;
+  bool lookup(ParmString, WordEntry &, const SensitiveCompare &) const;
 
   bool clean_lookup(ParmString sondslike, WordEntry &) const;
 
@@ -633,13 +556,12 @@ public:
   WordEntryEnumeration * detailed_elements() const;
   SoundslikeEnumeration * soundslike_elements() const;
       
-  PosibErr<void> add_repl(ParmString mis, ParmString cor) {
-    return Dictionary::add_repl(mis,cor);}
+  using Dictionary::add_repl;
   PosibErr<void> add_repl(ParmString mis, ParmString cor, ParmString s);
 
 private:
   PosibErr<void> save(FStream &, ParmString );
-  PosibErr<void> merge(FStream &, ParmString , Config * config);
+  PosibErr<void> merge(FStream &, ParmString , const Config * config);
 };
 
 WritableReplDict::Size WritableReplDict::size() const 
@@ -652,15 +574,15 @@ bool WritableReplDict::empty() const
   return word_lookup->empty();
 }
     
-bool WritableReplDict::lookup(ParmString word, const SensitiveCompare * c,
-                              WordEntry & o) const
+bool WritableReplDict::lookup(ParmString word, WordEntry & o,
+                           const SensitiveCompare & c) const
 {
   o.clear();
   pair<WordLookup::iterator, WordLookup::iterator> p(word_lookup->equal_range(word));
   while (p.first != p.second) {
-    if ((*c)(word,*p.first)) {
+    if (c(word,*p.first)) {
       o.what = WordEntry::Misspelled;
-      set_word(o, *p.first);
+      o.word = *p.first;
       o.intr[0] = (void *)*p.first;
       return true;
     }
@@ -675,7 +597,7 @@ bool WritableReplDict::clean_lookup(ParmString sl, WordEntry & o) const
   pair<WordLookup::iterator, WordLookup::iterator> p(word_lookup->equal_range(sl));
   if (p.first == p.second) return false;
   o.what = WordEntry::Misspelled;
-  set_word(o, *p.first);
+  o.word = *p.first;
   o.intr[0] = (void *)*p.first;
   return true;
   // FIXME: Deal with multiple entries
@@ -683,7 +605,7 @@ bool WritableReplDict::clean_lookup(ParmString sl, WordEntry & o) const
 
 bool WritableReplDict::soundslike_lookup(const WordEntry & word, WordEntry & o) const 
 {
-  if (use_soundslike) {
+  if (have_soundslike) {
     const StrVector * tmp = (const StrVector *)(word.intr[0]);
     o.clear();
     o.what = WordEntry::Misspelled;
@@ -691,15 +613,13 @@ bool WritableReplDict::soundslike_lookup(const WordEntry & word, WordEntry & o)
   } else {
     o.what = WordEntry::Misspelled;
     o.word = word.word;
-    o.word_size = word.word_size;
-    o.aff = "";
   }
   return true;
 }
 
 bool WritableReplDict::soundslike_lookup(ParmString soundslike, WordEntry & o) const
 {
-  if (use_soundslike) {
+  if (have_soundslike) {
     o.clear();
     SoundslikeLookup::const_iterator i = soundslike_lookup_.find(soundslike);
     if (i == soundslike_lookup_.end()) {
@@ -715,7 +635,7 @@ bool WritableReplDict::soundslike_lookup(ParmString soundslike, WordEntry & o) c
 }
 
 SoundslikeEnumeration * WritableReplDict::soundslike_elements() const {
-  if (use_soundslike)
+  if (have_soundslike)
     return new SoundslikeElements(soundslike_lookup_.begin(), 
                                   soundslike_lookup_.end());
   else
@@ -730,9 +650,9 @@ WritableReplDict::Enum * WritableReplDict::detailed_elements() const {
 
 static void repl_next(WordEntry * w)
 {
-  const Str * & i   = (const Str * &)(w->intr[0]);
-  const Str *   end = (const Str *  )(w->intr[1]);
-  set_word(*w, *i);
+  const Str * i   = (const Str *)(w->intr[0]);
+  const Str * end = (const Str *)(w->intr[1]);
+  w->word = *i;
   ++i;
   if (i == end) w->adv_ = 0;
 }
@@ -742,7 +662,8 @@ static void repl_init(const StrVector * tmp, WordEntry & o)
   o.what = WordEntry::Word;
   const Str * i   = tmp->pbegin();
   const Str * end = tmp->pend();
-  set_word(o, *i);
+  o.word = *i;
+  o.aff  = "";
   ++i;
   if (i != end) {
     o.intr[0] = (void *)i;
@@ -761,7 +682,7 @@ bool WritableReplDict::repl_lookup(const WordEntry & w, WordEntry & o) const
   } else {
     SensitiveCompare c(lang()); // FIXME: This is not exactly right
     WordEntry tmp;
-    WritableReplDict::lookup(w.word, &c, tmp);
+    WritableReplDict::lookup(w.word, tmp, c);
     repls = get_vector(tmp.word);
     if (!repls) return false;
   }
@@ -779,7 +700,7 @@ bool WritableReplDict::repl_lookup(ParmString word, WordEntry & o) const
 
 PosibErr<void> WritableReplDict::add_repl(ParmString mis, ParmString cor, ParmString sl) 
 {
-  Str m;
+  Str m, c, s;
   SensitiveCompare cmp(lang()); // FIXME: I don't think this is completely correct
   WordEntry we;
 
@@ -789,13 +710,11 @@ PosibErr<void> WritableReplDict::add_repl(ParmString mis, ParmString cor, ParmSt
   for (; p != p0.second && !cmp(mis,*p); ++p);
 
   if (p == p0.second) {
-    byte * m0  = (byte *)buffer.alloc(sizeof(StrVector) + mis.size() + 3, sizeof(void *));
+    char * m0  = (char *)buffer.alloc(sizeof(StrVector) + mis.size() + 1, sizeof(void *));
     new (m0) StrVector;
     m0 += sizeof(StrVector);
-    *m0++ = lang()->get_word_info(mis);
-    *m0++ = mis.size();
     memcpy(m0, mis.str(), mis.size() + 1);
-    m = (char *)m0;
+    m = m0;
     p = word_lookup->insert(m).first;
   } else {
     m = *p;
@@ -806,17 +725,12 @@ PosibErr<void> WritableReplDict::add_repl(ParmString mis, ParmString cor, ParmSt
   for (StrVector::iterator i = v->begin(); i != v->end(); ++i)
     if (cmp(cor, *i)) return no_err;
     
-  byte * c0 = (byte *)buffer.alloc(cor.size() + 3);
-  *c0++ = lang()->get_word_info(cor);
-  *c0++ = cor.size();
-  memcpy(c0, cor.str(), cor.size() + 1);
-  v->push_back((char *)c0);
+  c = buffer.dup(cor);
+  get_vector(m)->push_back(c);
 
-  if (use_soundslike) {
-    byte * s0 = (byte *)buffer.alloc(sl.size() + 2);
-    *s0++ = sl.size();
-    memcpy(s0, sl.str(), sl.size() + 1);
-    soundslike_lookup_[(char *)s0].push_back(m);
+  if (have_soundslike) {
+    s = buffer.dup(sl);
+    soundslike_lookup_[s].push_back(m);
   }
 
   return no_err;
@@ -837,21 +751,19 @@ PosibErr<void> WritableReplDict::save (FStream & out, ParmString file_name)
     StrVector * v = get_vector(*i);
     for (StrVector::iterator j = v->begin(); j != v->end(); ++j)
     {
-      write_n_escape(out, conv1(*i));
-      out << ' ';
-      write_n_escape(out, conv2(*j));
-      out << '\n';
+      out.printf("%s %s\n", conv1(*i), conv2(*j));
     }
   }
   return no_err;
 }
 
 PosibErr<void> WritableReplDict::merge(FStream & in,
-                                       ParmString file_name, 
-                                       Config * config)
+                                    ParmString file_name, 
+                                    const Config * config)
 {
   typedef PosibErr<void> Ret;
   unsigned int version;
+  String word, mis, sound, repl;
   unsigned int num_words, num_repls;
 
   String buf;
@@ -870,7 +782,7 @@ PosibErr<void> WritableReplDict::merge(FStream & in,
 
   split(dp);
   {
-    Ret pe = set_check_lang(dp.key, *config);
+    Ret pe = set_check_lang(dp.key, config);
     if (pe.has_err())
       return pe.with_file(file_name);
   }
@@ -885,29 +797,24 @@ PosibErr<void> WritableReplDict::merge(FStream & in,
 
   split(dp);
   if (dp.key.size > 0)
-    set_file_encoding(dp.key, *config);
+    set_file_encoding(dp.key, config);
   else
-    set_file_encoding("", *config);
+    set_file_encoding("", config);
 
   if (version == 11) {
 
     ConvP conv1(iconv);
     ConvP conv2(iconv);
-    for (;;) {
-      bool res = getline_n_unescape(in, buf, '\n');
-      if (!res) break;
-      char * mis = buf.mstr();
-      char * repl = strchr(mis, ' ');
-      if (!repl) continue; // bad line, ignore
-      *repl = '\0'; // split string
-      ++repl;
-      if (!repl[0]) continue; // empty repl, ignore
+    do {
+      in.getline(mis, ' ');
+      if (!in) break;
+      in.getline(repl, '\n');
+      if (!in) make_err(bad_file_format, file_name);
       WritableReplDict::add_repl(conv1(mis), conv2(repl));
-    }
-    
+    } while (true);
+
   } else {
-    
-    String mis, sound, repl;
+
     unsigned int h,i,j;
     for (h=0; h != num_soundslikes; ++h) {
       in >> sound >> num_words;
@@ -938,12 +845,12 @@ WritableReplDict::~WritableReplDict()
 
 namespace aspeller {
 
-  Dictionary * new_default_writable_dict(const Config & cfg) {
-    return new WritableDict(cfg);
+  Dictionary * new_default_writable_dict() {
+    return new WritableDict();
   }
 
-  Dictionary * new_default_replacement_dict(const Config & cfg) {
-    return new WritableReplDict(cfg);
+  Dictionary * new_default_replacement_dict() {
+    return new WritableReplDict();
   }
 
 }
